import unittest
from ..oocana.throttler import throttle
import time
class TestThrottle(unittest.TestCase):
    
    def test_throttle(self):
        
        a = 0

        @throttle(1)
        def test_fn(i):
            nonlocal a
            a = i

        self.assertEqual(a, 0)
        test_fn(1)
        self.assertEqual(a, 1)

        test_fn(2)
        self.assertEqual(a, 1)
        test_fn(3)
        self.assertEqual(a, 1)
        
        time.sleep(1.1)
        self.assertEqual(a, 3)

        test_fn(4)
        test_fn(5)
        test_fn(6)

        time.sleep(1.1)
        self.assertEqual(a, 6)

        test_fn(7)
        time.sleep(1.1)
        self.assertEqual(a, 7)


from .data import *  # noqa: F403
from .context import *  # noqa: F403
from .service import *  # noqa: F403
from .handle_data import *  # noqa: F403
from .preview import *  # noqa: F403
from .extra import *  # noqa: F403
from .schema import *  # noqa: F403
from .mainframe import Mainframe as Mainframe  # noqa: F403



from typing import Literal, TypedDict

__all__ = ["LLMModelOptions", "LLMMessage"]

class LLMModelOptions(TypedDict):
    model: str
    temperature: float
    top_p: float
    max_tokens: int

class LLMMessage(TypedDict):
    role: Literal["system", "user", "assistant"]
    content: str



from dataclasses import dataclass
from typing import Any, Optional
from .schema import FieldSchema, ContentMediaType

__all__ = ["HandleDef", "InputHandleDef"]

@dataclass(frozen=True, kw_only=True)
class HandleDef:
    """The base handle for output def, can be directly used for output def 
    """
    handle: str
    """The name of the handle. it should be unique in handle list."""

    """The description of the handle. It is used to display in the UI."""
    description: Optional[str] = None

    json_schema: Optional[FieldSchema] = None
    """The schema of the handle. It is used to validate the handle's content."""

    kind: Optional[str] = None
    """A alias of the handle's type name. It is used to display in the UI and connect to the other handle match"""

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)
        if "handle" not in kwargs:
            raise ValueError("missing attr key: 'handle'")
        json_schema = self.json_schema
        if json_schema is not None and not isinstance(json_schema, FieldSchema):
            object.__setattr__(self, "json_schema", FieldSchema.generate_schema(json_schema))

    def check_handle_type(self, type: ContentMediaType) -> bool:
        if self.handle is None:
            return False
        if self.json_schema is None:
            return False
        if self.json_schema.contentMediaType is None:
            return False
        return self.json_schema.contentMediaType == type

    def is_var_handle(self) -> bool:
        return self.check_handle_type("oomol/var")
    
    def is_secret_handle(self) -> bool:
        return self.check_handle_type("oomol/secret")

    def is_bin_handle(self) -> bool:
        return self.check_handle_type("oomol/bin")

@dataclass(frozen=True, kw_only=True)
class InputHandleDef(HandleDef):

    value: Optional[Any] = None
    """default value for input handle, can be None.
    """

    nullable: bool = False
    """If the handle value can be None. If True, the handle can be None, otherwise it must have a value."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


from dataclasses import asdict
from json import loads
from .data import BlockInfo, StoreKey, JobDict, BlockDict, BinValueDict, VarValueDict
from .handle_data import HandleDef
from .mainframe import Mainframe
from typing import Dict, Any, TypedDict, Optional
from base64 import b64encode
from io import BytesIO
from .throttler import throttle
from .preview import PreviewPayload, TablePreviewData, DataFrame, ShapeDataFrame, PartialDataFrame
from .data import EXECUTOR_NAME
import os.path
import logging
import copy

__all__ = ["Context", "HandleDefDict"]

class HandleDefDict(TypedDict):
    """a dict that represents the handle definition, used in the block schema output and input defs.
    """

    handle: str
    """the handle of the output, should be defined in the block schema output defs, the field name is handle
    """

    description: str | None
    """the description of the output, should be defined in the block schema output defs, the field name is description
    """

    json_schema: Dict[str, Any] | None
    """the schema of the output, should be defined in the block schema output defs, the field name is json_schema
    """

    kind: str | None
    """the kind of the output, should be defined in the block schema output defs, the field name is kind
    """

    nullable: bool
    """if the output can be None, should be defined in the block schema output defs, the field name is nullable
    """

class OnlyEqualSelf:
    def __eq__(self, value: object) -> bool:
        return self is value

class OOMOL_LLM_ENV(TypedDict):
    base_url: str
    """{basUrl}/v1 openai compatible endpoint
    """
    base_url_v1: str
    api_key: str
    models: list[str]

class HostInfo(TypedDict):
    gpu_vendor: str
    gpu_renderer: str

class Context:
    __inputs: Dict[str, Any]

    __block_info: BlockInfo
    __outputs_def: Dict[str, HandleDef]
    __inputs_def: Dict[str, Any]
    __store: Any
    __keep_alive: OnlyEqualSelf = OnlyEqualSelf()
    __session_dir: str
    __tmp_dir: str
    __package_name: str | None = None
    _logger: Optional[logging.Logger] = None
    __pkg_dir: str

    def __init__(
        self, *, inputs: Dict[str, Any], blockInfo: BlockInfo, mainframe: Mainframe, store, inputs_def, outputs_def, session_dir: str, tmp_dir: str, package_name: str, pkg_dir: str
    ) -> None:

        self.__block_info = blockInfo

        self.__mainframe = mainframe
        self.__store = store
        self.__inputs = inputs

        outputs_defs = {}
        if outputs_def is not None:
            for k, v in outputs_def.items():
                outputs_defs[k] = HandleDef(**v)
        self.__outputs_def = outputs_defs
        self.__inputs_def = inputs_def
        self.__session_dir = session_dir
        self.__tmp_dir = tmp_dir
        self.__package_name = package_name
        self.__pkg_dir = pkg_dir

    @property
    def logger(self) -> logging.Logger:
        """a custom logger for the block, you can use it to log the message to the block log. this logger will report the log by context report_logger api.
        """

        # setup after init, so the logger always exists
        if self._logger is None:
            raise ValueError("logger is not setup, please setup the logger in the block init function.")
        return self._logger

    @property
    def session_dir(self) -> str:
        """a temporary directory for the current session, all blocks in the one session will share the same directory.
        """
        return self.__session_dir
    
    @property
    def tmp_dir(self) -> str:
        """a temporary directory for the current follow, all blocks in the this flow will share the same directory. this directory will be cleaned if this session finish successfully, otherwise it will be kept for debugging or other purpose.
        """
        return self.__tmp_dir
    
    @property
    def tmp_pkg_dir(self) -> str:
        """a temporary directory for the current package, all blocks in the this package will share the same directory. this directory will be cleaned if this session finish successfully, otherwise it will be kept for debugging or other purpose.
        """
        return os.path.join(self.__tmp_dir, self.__package_name) if self.__package_name else self.__tmp_dir

    @property
    def pkg_dir(self) -> str:
        """a directory for the current package, all blocks in the this package will share the same directory. this directory will be cleaned if this session finish successfully, otherwise it will be kept for debugging or other purpose.
        """
        return self.__pkg_dir

    @property
    def keepAlive(self):
        return self.__keep_alive

    @property
    def inputs(self):
        return self.__inputs
    
    @property
    def inputs_def(self) -> Dict[str, HandleDefDict]:
        return copy.deepcopy(self.__inputs_def) if self.__inputs_def is not None else {}

    @property
    def outputs_def(self) -> Dict[str, HandleDefDict]:
        outputs = {}
        for k, v in self.__outputs_def.items():
            outputs[k] = asdict(v)
        return outputs

    @property
    def session_id(self):
        return self.__block_info.session_id

    @property
    def job_id(self):
        return self.__block_info.job_id
    
    @property
    def job_info(self) -> JobDict:
        return self.__block_info.job_info()
    
    @property
    def block_info(self) -> BlockDict:
        return self.__block_info.block_dict()
    
    @property
    def node_id(self) -> str:
        return self.__block_info.stacks[-1].get("node_id", None)
    
    @property
    def oomol_llm_env(self) -> OOMOL_LLM_ENV:
        """this is a dict contains the oomol llm environment variables
        """
        return {
            "base_url": os.getenv("OOMOL_LLM_BASE_URL", ""),
            "base_url_v1": os.getenv("OOMOL_LLM_BASE_URL_V1", ""),
            "api_key": os.getenv("OOMOL_LLM_API_KEY", ""),
            "models": os.getenv("OOMOL_LLM_MODELS", "").split(","),
        }

    @property
    def host_info(self) -> HostInfo:
        """this is a dict contains the host information
        """
        return {
            "gpu_vendor": os.getenv("OOMOL_HOST_GPU_VENDOR", "unknown"),
            "gpu_renderer": os.getenv("OOMOL_HOST_GPU_RENDERER", "unknown"),
        }

    @property
    def host_endpoint(self) -> str | None:
        """A host endpoint that allows containers to access services running on the host system.
        
        Returns:
            str: The host endpoint if available.
            None: If the application is running in a cloud environment where no host endpoint is defined.
        """
        return os.getenv("OO_HOST_ENDPOINT", None)

    def __store_ref(self, handle: str):
        return StoreKey(
            executor=EXECUTOR_NAME,
            handle=handle,
            job_id=self.job_id,
            session_id=self.session_id,
        )
    
    def __is_basic_type(self, value: Any) -> bool:
        return isinstance(value, (int, float, str, bool))
    
    def __wrap_output_value(self, handle: str, value: Any):
        """
        wrap the output value:
        if the value is a var handle, store it in the store and return the reference.
        if the value is a bin handle, store it in the store and return the reference.
        if the handle is not defined in the block outputs schema, raise an ValueError.
        otherwise, return the value.
        :param handle: the handle of the output
        :param value: the value of the output
        :return: the wrapped value
        """
        # __outputs_def should never be None
        if self.__outputs_def is None:
            return value
        
        output_def = self.__outputs_def.get(handle)
        if output_def is None:
            raise ValueError(
                f"Output handle key: [{handle}] is not defined in Block outputs schema."
            )
        
        if output_def.is_var_handle() and not self.__is_basic_type(value):
            ref = self.__store_ref(handle)
            self.__store[ref] = value
            var: VarValueDict = {
                "__OOMOL_TYPE__": "oomol/var",
                "value": asdict(ref)
            }
            return var
        
        if output_def.is_bin_handle():
            if not isinstance(value, bytes):
                self.send_warning(
                    f"Output handle key: [{handle}] is defined as binary, but the value is not bytes."
                )
                return value
            
            bin_file = f"{self.session_dir}/binary/{self.session_id}/{self.job_id}/{handle}"
            os.makedirs(os.path.dirname(bin_file), exist_ok=True)
            try:
                with open(bin_file, "wb") as f:
                    f.write(value)
            except IOError as e:
                raise IOError(
                    f"Output handle key: [{handle}] is defined as binary, but an error occurred while writing the file: {e}"
                )

            if os.path.exists(bin_file):
                bin_value: BinValueDict = {
                    "__OOMOL_TYPE__": "oomol/bin",
                    "value": bin_file,
                }
                return bin_value
            else:
                raise IOError(
                    f"Output handle key: [{handle}] is defined as binary, but the file is not written."
                )
        return value

    def output(self, key: str, value: Any):
        """
        output the value to the next block

        key: str, the key of the output, should be defined in the block schema output defs, the field name is handle
        value: Any, the value of the output
        """

        try:
            wrap_value = self.__wrap_output_value(key, value)
        except ValueError as e:
            self.send_warning(
                f"{e}"
            )
            return
        except IOError as e:
            self.send_warning(
                f"{e}"
            )
            return

        node_result = {
            "type": "BlockOutput",
            "handle": key,
            "output": wrap_value,
        }
        self.__mainframe.send(self.job_info, node_result)
    
    def outputs(self, outputs: Dict[str, Any]):
        """
        output the value to the next block

        map: Dict[str, Any], the key of the output, should be defined in the block schema output defs, the field name is handle
        """

        values = {}
        for key, value in outputs.items():
            try:
                wrap_value = self.__wrap_output_value(key, value)
                values[key] = wrap_value
            except ValueError as e:
                self.send_warning(
                    f"{e}"
                )
            except IOError as e:
                self.send_warning(
                    f"{e}"
                )
        self.__mainframe.send(self.job_info, {
            "type": "BlockOutputs",
            "outputs": values,
        })

        

    def finish(self, *, result: Dict[str, Any] | None = None, error: str | None = None):
        """
        finish the block, and send the result to oocana.
        if error is not None, the block will be finished with error.
        then if result is not None, the block will be finished with result.
        lastly, if both error and result are None, the block will be finished without any result.
        """

        if error is not None:
            self.__mainframe.send(self.job_info, {"type": "BlockFinished", "error": error})
        elif result is not None:
            wrap_result = {}
            if isinstance(result, dict):
                for key, value in result.items():
                    try:
                        wrap_result[key] = self.__wrap_output_value(key, value)
                    except ValueError as e:
                        self.send_warning(
                            f"Output handle key: [{key}] is not defined in Block outputs schema. {e}"
                        )
                    except IOError as e:
                        self.send_warning(
                            f"Output handle key: [{key}] is not defined in Block outputs schema. {e}"
                        )

                self.__mainframe.send(self.job_info, {"type": "BlockFinished", "result": wrap_result})
            else:
                raise ValueError(
                    f"result should be a dict, but got {type(result)}"
                )
        else:
            self.__mainframe.send(self.job_info, {"type": "BlockFinished"})

    def send_message(self, payload):
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockMessage",
                "payload": payload,
            },
        )
    
    def __dataframe(self, payload: PreviewPayload) -> PreviewPayload:
        def not_default_index(df: DataFrame) -> bool:
            index = df.index.tolist()
            default_index = list(range(len(df)))
            return index != default_index

        if isinstance(payload, DataFrame):
            payload = { "type": "table", "data": payload }

        if isinstance(payload, dict) and payload.get("type") == "table":
            df = payload.get("data")
            if isinstance(df, ShapeDataFrame):
                row_count = df.shape[0]
                if row_count <= 10:
                    data = loads(df.to_json(orient='split'))
                    columns = data.get("columns", [])
                    rows = data.get("data", [])
                    if not_default_index(df):
                        index = df.index.tolist()
                        rows = [[index[i], *rows[i]] for i in range(len(rows))]
                        columns = ["", *columns]

                elif isinstance(df, PartialDataFrame):
                    need_add_index = False # TODO: some index is not begin with 0, current just always hide index
                    # to_json will serialize some default json dumps not supported type like datetime, so we just use to_json for now.
                    head_data = loads(df.head(5).to_json(orient='split'))
                    columns = head_data.get("columns", [])

                    rows_head = head_data.get("data", [])
                    if need_add_index:
                        columns = ["", *columns]
                        head_index = head_data.get("index", [])
                        rows_head = [[head_index[i], *rows_head[i]] for i in range(len(rows_head))]
                    
                    tail_data = loads(df.tail(5).to_json(orient='split'))
                    rows_tail = tail_data.get("data", [])
                    if need_add_index:
                        tail_index = tail_data.get("index", [])
                        rows_tail = [[tail_index[i], *rows_tail[i]] for i in range(len(rows_tail))]

                    rows_dots = [["..."] * len(columns)]
                    rows = rows_head + rows_dots + rows_tail
                else:
                    print("dataframe more than 10 rows but not support head and tail is not supported")
                    return payload
                data: TablePreviewData = { "rows": rows, "columns": columns, "row_count": row_count }
                payload = { "type": "table", "data": data }
            else:
                print("dataframe is not support shape property")
        
        return payload

    def __matplotlib(self, payload: PreviewPayload) -> PreviewPayload:
        # payload is a matplotlib Figure
        if hasattr(payload, 'savefig'):
            fig: Any = payload
            buffer = BytesIO()
            fig.savefig(buffer, format='png')
            buffer.seek(0)
            png = buffer.getvalue()
            buffer.close()
            url = f'data:image/png;base64,{b64encode(png).decode("utf-8")}'
            payload = { "type": "image", "data": url }

        return payload

    def preview(self, payload: PreviewPayload):
        payload = self.__dataframe(payload)
        payload = self.__matplotlib(payload)

        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockPreview",
                "payload": payload,
            },
        )

    @throttle(0.3)
    def report_progress(self, progress: float | int):
        """report progress

        This api is used to report the progress of the block. but it just effect the ui progress not the real progress.
        This api is throttled. the minimum interval is 0.3s. 
        When you first call this api, it will report the progress immediately. After it invoked once, it will report the progress at the end of the throttling period.

        |       0.25 s        |   0.2 s  |
        first call       second call    third call  4 5 6 7's calls
        |                     |          |          | | | |
        | -------- 0.3 s -------- | -------- 0.3 s -------- |
        invoke                  invoke                    invoke
        :param float | int progress: the progress of the block, the value should be in [0, 100].
        """
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockProgress",
                "rate": progress,
            }
        )

    def report_log(self, line: str, stdio: str = "stdout"):
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockLog",
                "log": line,
                stdio: stdio,
            },
        )

    def log_json(self, payload):
        self.__mainframe.report(
            self.block_info,
            {
                "type": "BlockLogJSON",
                "json": payload,
            },
        )

    def send_warning(self, warning: str):
        self.__mainframe.report(self.block_info, {"type": "BlockWarning", "warning": warning})

    def send_error(self, error: str):
        '''
        deprecated, use error(error) instead.
        consider to remove in the future.
        '''
        self.error(error)

    def error(self, error: str):
        self.__mainframe.send(self.job_info, {"type": "BlockError", "error": error})



import unittest
from oocana import Mainframe
from oocana import JobDict
class TestMainframe(unittest.TestCase):

  __mainframe: Mainframe
  __job_info: JobDict = {
    #  broker 里对这两个字段的 send 做了特殊兼容，需要固定值。
    'session_id': '123',
    'job_id': '123',
  }

  def setUp(self):
    self.__mainframe = Mainframe('mqtt://localhost:47688')
    self.__mainframe.connect()
  
  def tearDown(self) -> None:
     assert self.__mainframe is not None
     self.__mainframe.disconnect()
     return super().tearDown()

  # def test_send(self):
  #   info = self.__mainframe.send(self.__job_info, {
  #     'dir': '123',
  #   })

  #   info.wait_for_publish()

  # def test_send_ready(self):
  #   self.__mainframe.notify_ready('123', '123')

# 激活虚拟环境后，执行以下命令：
# python -m unittest oocana/tests/mainframe_test.py
# python -m unittest oocana.tests.mainframe_test.TestMainframe.[单个方法名]
if __name__ == '__main__':
    unittest.main()



from typing import Literal, Dict, Optional, TypeAlias, Any, cast
from dataclasses import dataclass
from .data import BinValueDict, VarValueDict

__all__ = ["FieldSchema", "PrimitiveFieldSchema", "VarFieldSchema", "SecretFieldSchema", "ArrayFieldSchema", "ObjectFieldSchema", "ContentMediaType", "is_bin_value", "is_var_value", "is_array_dict", "is_object_dict", "is_var_dict", "is_primitive_dict", "is_secret_dict"]

OomolType = Literal["oomol/var", "oomol/secret", "oomol/bin"]

ContentMediaType: TypeAlias = Literal["oomol/bin", "oomol/secret", "oomol/var"]

def is_bin_value(d: BinValueDict | Any):
    if isinstance(d, dict) is False:
        return False
    dd = cast(BinValueDict, d)
    return dd.get("__OOMOL_TYPE__") == "oomol/bin" and isinstance(dd.get("value") , str) 

def is_var_value(d: VarValueDict | Any):
    if isinstance(d, dict) is False:
        return False
    dd = cast(VarValueDict, d)
    return dd.get("__OOMOL_TYPE__") == "oomol/var" and isinstance(dd.get("value") , dict)

def is_array_dict(dict: Dict):
    return dict.get("type") == "array"

def is_object_dict(dict: Dict):
    return dict.get("type") == "object"

def is_var_dict(dict: Dict):
    return dict.get("contentMediaType") == "oomol/var"

def is_primitive_dict(dict: Dict):
    return dict.get("type") in ["string", "number", "boolean"] and dict.get("contentMediaType") is None

def is_secret_dict(dict: Dict):
    return dict.get("contentMediaType") == "oomol/secret" and dict.get("type") == "string"

@dataclass(frozen=True, kw_only=True)
class FieldSchema:
    """ The JSON schema of the handle. It contains the schema of the handle's content.
        but we only need the contentMediaType to check the handle's type here.
    """

    contentMediaType: Optional[ContentMediaType] = None
    """The media type of the content of the schema."""

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

    @staticmethod
    def generate_schema(dict: Dict):
        if is_var_dict(dict):
            return VarFieldSchema(**dict)
        elif is_secret_dict(dict):
            return SecretFieldSchema(**dict)
        elif is_primitive_dict(dict):
            return PrimitiveFieldSchema(**dict)
        elif is_array_dict(dict):
            return ArrayFieldSchema(**dict)
        elif is_object_dict(dict):
            return ObjectFieldSchema(**dict)
        else:
            return FieldSchema(**dict)

@dataclass(frozen=True, kw_only=True)
class PrimitiveFieldSchema(FieldSchema):
    type: Literal["string", "number", "boolean"]
    contentMediaType: None = None
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)


@dataclass(frozen=True, kw_only=True)
class VarFieldSchema(FieldSchema):
    contentMediaType: Literal["oomol/var"] = "oomol/var"

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

@dataclass(frozen=True, kw_only=True)
class SecretFieldSchema(FieldSchema):
    type: Literal["string"] = "string"
    contentMediaType: Literal["oomol/secret"] = "oomol/secret"

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

@dataclass(frozen=True, kw_only=True)
class ArrayFieldSchema(FieldSchema):
    type: Literal["array"] = "array"
    items: Optional['FieldSchema'] = None

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)
        items = self.items
        if items is not None and not isinstance(items, FieldSchema):
            object.__setattr__(self, "items", FieldSchema.generate_schema(items))

@dataclass(frozen=True, kw_only=True)
class ObjectFieldSchema(FieldSchema):
    type: Literal["object"] = "object"
    properties: Optional[Dict[str, 'FieldSchema']] = None

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

        properties = self.properties
        if properties is not None:
            for key, value in properties.items():
                if not isinstance(value, FieldSchema):
                    properties[key] = FieldSchema.generate_schema(value)




from typing import Literal, Callable, Any, TypedDict, Optional, TypeAlias, Union
from .context import Context
from .data import JobDict
from abc import ABC, abstractmethod

__all__ = ["ServiceMessage", "BlockHandler", "ServiceContextAbstractClass", "ServiceExecutor", "ServiceExecutePayload", "StopAtOption"]

class ServiceMessage(TypedDict):
    job_id: str
    node_id: str
    flow_path: str
    payload: Any

BlockHandler: TypeAlias = Union[Callable[[str, Any, Context], Any], dict[str, Callable[[Any, Context], Any]]]

class ServiceContextAbstractClass(ABC):
    
    @property
    @abstractmethod
    def block_handler(self) -> BlockHandler:
        pass
    
    @block_handler.setter
    @abstractmethod
    def block_handler(self, value: BlockHandler):
        pass

    def __setitem__(self, key: str, value: Any):
        pass
    
    @property
    @abstractmethod
    def waiting_ready_notify(self) -> bool:
        """set to True if the service need to wait for the ready signal before start. default is false which means the service will start immediately after block_handler is set.
            this function need to be called before set block_handler
            after set this function, developer need call notify_ready manually when the service is ready to start, otherwise the service will not run block
        """
        pass

    @waiting_ready_notify.setter
    @abstractmethod
    def waiting_ready_notify(self, value: bool):
        pass

    def notify_ready(self):
        """notify the service that the service is ready to start. this function need to be called after waiting_ready_notify is set to True otherwise this function has no effect"""
        pass

    def add_message_callback(self, callback: Callable[[ServiceMessage], Any]):
        """add a callback to handle the message to the service, the callback will be called when the message is sent to the service
        :param callback: the callback to handle the message
        """
        pass

StopAtOption: TypeAlias = Optional[Literal["block_end", "session_end", "app_end", "never"]]

class ServiceExecutor(TypedDict):
    name: str
    entry: str
    function: str
    start_at: Optional[Literal["block_start", "session_start", "app_start"]]
    stop_at: StopAtOption
    keep_alive: Optional[int]

class ServiceExecutePayload(JobDict):
    dir: str
    block_name: str
    service_executor: ServiceExecutor
    outputs: dict
    service_hash: str


import unittest
from oocana import FieldSchema, is_secret_dict, SecretFieldSchema, is_var_dict, VarFieldSchema, is_array_dict, ArrayFieldSchema, is_object_dict, ObjectFieldSchema
from typing import cast

secret_fixture = {
    "type": "string",
    "contentMediaType": "oomol/secret"
}

var_fixture = {
    "contentMediaType": "oomol/var"
}

array_fixture = {
    "type": "array",
    "items": {
        "type": "string"
    }
}

array_secret_fixture = {
    "type": "array",
    "items": {
        "type": "string",
        "contentMediaType": "oomol/secret"
    }
}

object_fixture = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string"
        }
    }
}

object_secret_fixture = {
    "type": "object",
    "properties": {
        "name": {
            "type": "string",
        },
        "password": {
            "type": "string",
            "contentMediaType": "oomol/secret"
        }
    }
}

class TestSchema(unittest.TestCase):
    def test_secret_field_schema(self):
        self.assertTrue(is_secret_dict(secret_fixture))
        secret = FieldSchema.generate_schema(secret_fixture)
        self.assertIsInstance(secret, SecretFieldSchema)

    def test_var_field_schema(self):
        self.assertTrue(is_var_dict(var_fixture))
        var = FieldSchema.generate_schema(var_fixture)
        self.assertIsInstance(var, VarFieldSchema)
    
    def test_array_field_schema(self):
        self.assertTrue(is_array_dict(array_fixture))
        array = FieldSchema.generate_schema(array_fixture)
        self.assertIsInstance(array, ArrayFieldSchema)

    def test_object_field_schema(self):
        self.assertTrue(is_object_dict(object_fixture))
        obj = FieldSchema.generate_schema(object_fixture)
        self.assertIsInstance(obj, ObjectFieldSchema)

    def test_array_secret_field_schema(self):
        self.assertTrue(is_array_dict(array_secret_fixture))
        array_secret = FieldSchema.generate_schema(array_secret_fixture)
        self.assertIsInstance(array_secret, ArrayFieldSchema)

        array_secret = cast(ArrayFieldSchema, array_secret)
        self.assertIsInstance(array_secret.items, SecretFieldSchema)

    def test_object_secret_field_schema(self):
        self.assertTrue(is_object_dict(object_secret_fixture))
        obj_secret = FieldSchema.generate_schema(object_secret_fixture)
        self.assertIsInstance(obj_secret, ObjectFieldSchema)

        obj_secret = cast(ObjectFieldSchema, obj_secret)
        self.assertIsNotNone(obj_secret.properties)

        if obj_secret.properties is not None:
            self.assertIsInstance(obj_secret.properties["password"], SecretFieldSchema)
        else:
            self.fail("properties is None")


import unittest
from oocana import handle_data
from typing import cast


fixture = {
    "handle": "test",
    "json_schema": {
        "contentMediaType": "oomol/bin"
    },
    "name": "options"
}

class TestHandleData(unittest.TestCase):
    def test_handle_def_field(self):

        d = fixture.copy()
        handle_data.HandleDef(**d)

        d = fixture.copy()
        del d["name"]

        handle_def = handle_data.HandleDef(**d)
        self.assertEqual(handle_def.handle, "test")
        self.assertIsNotNone(handle_def.json_schema)
        
        json_schema = cast(handle_data.FieldSchema, handle_def.json_schema)
        self.assertEqual(json_schema.contentMediaType, "oomol/bin")

        d = fixture.copy()
        del d["name"]
        del d["json_schema"]

        handle_def = handle_data.HandleDef(**d)
        self.assertEqual(handle_def.handle, "test")
        self.assertIsNone(handle_def.json_schema)

    def test_handle_def_extra_field(self):
        d = fixture.copy()
        d["a"] = "a"

        handle_def = handle_data.HandleDef(**d)
        self.assertEqual(handle_def.handle, "test")

    def test_handle_def_missing_field(self):
        d = {
            "a": "1",
        }

        with self.assertRaises(ValueError, msg="missing attr key: 'handle'"):
            handle_data.InputHandleDef(**d) # type: ignore

    def test_handle_type(self):
        d = fixture.copy()
        d["json_schema"]["contentMediaType"] = "oomol/secret"

        handle_def = handle_data.HandleDef(**d)
        self.assertTrue(handle_def.is_secret_handle())

        d["json_schema"]["contentMediaType"] = "oomol/var"
        handle_def = handle_data.HandleDef(**d)
        self.assertTrue(handle_def.is_var_handle())

        d = {
            "handle": "auto_slices",
            "json_schema": {
                "items": {
                    "properties": {
                        "begin": { "type": "number" },
                        "end": { "type": "number" }
                    },
                    "required": ["begin", "end"],
                    "type": "object"
                },
                "type": "array"
            }
        }
        handle_def = handle_data.HandleDef(**d)
        self.assertFalse(handle_def.is_var_handle())
        self.assertFalse(handle_def.is_secret_handle())

    def test_input_handle_type(self):
        d = fixture.copy()
        d["json_schema"]["contentMediaType"] = "oomol/secret"

        handle_def = handle_data.InputHandleDef(**d)
        self.assertTrue(handle_def.is_secret_handle())

        d["json_schema"]["contentMediaType"] = "oomol/var"
        handle_def = handle_data.InputHandleDef(**d)
        self.assertTrue(handle_def.is_var_handle())

        d = {
            "handle": "auto_slices",
            "json_schema": {
                "items": {
                    "properties": {
                        "begin": { "type": "number" },
                        "end": { "type": "number" }
                    },
                    "required": ["begin", "end"],
                    "type": "object"
                    },
                "type": "array"
            }
        }
        handle_def = handle_data.InputHandleDef(**d)
        self.assertFalse(handle_def.is_var_handle())
        self.assertFalse(handle_def.is_secret_handle())